﻿using UnityEngine;
using System.Collections;
using UnityEngine.Events;
namespace com.geeitsomelaldy.platforming
{ 
    [RequireComponent(typeof(BaseCollider))]
    public abstract class LaldyPlatformBase : MonoBehaviour
    {
        public Vector2 Velocity;
        public Vector2 MaximumVelocity;

        /// <summary>
        /// Size of the Platformer Collider, used to detect ground, ceiling and wall hits
        /// </summary>
        //public Rect ColliderBounds;
        private BaseCollider _baseCollider;
        protected Vector2 ColliderSize
        {
            get
            {
                if (_baseCollider == null)
                {
                    _baseCollider = this.GetComponent<BaseCollider>();
                }
                return _baseCollider.HitBounds.size;
            }
        }
        protected Vector2 ColliderOffset
        {
            get
            {
                if (_baseCollider == null)
                {
                    _baseCollider = this.GetComponent<BaseCollider>();
                }
                return _baseCollider.HitBounds.position;
            }
        }

        /// <summary>
        /// 3 raycasts are used to detect ground, 
        /// one at (0,0), 
        /// one at (ColliderSize.x*DownTestFactor,0) 
        /// and one at (-ColliderSize.x * DownTestfactor,0)
        /// </summary>
        public float DownTestFactor = 0.75f;
        public void CollidersDirty()
        {
            _downTestOffsets = null;
            _forwardTestOffsets = null;
            DownTest(1.5f);
        }
        private Vector3[] _downTestOffsets;
        private Vector3[] DownTestOffsets
        {
            get
            {
                if (_downTestOffsets == null)
                {
                    if (DownTestFactor > 0)
                    { 
                        _downTestOffsets = new Vector3[] {
                            new Vector3(0,0,0),
                            new Vector3(DownTestFactor * ColliderSize.x,0,0),
                            new Vector3(-DownTestFactor * ColliderSize.x,0,0),
                        };
                    }
                    else
                    {
                        _downTestOffsets = originOffsets;
                    }
                }
                return _downTestOffsets;
            }
        }

        /// <summary>
        /// 3 raycasts are used to detect wall collisions, 
        /// one at (0,0), 
        /// one at (0,ColliderSize.y*ForwardTestFactor) 
        /// and one at (0,-ColliderSize.y * ForwardTestFactor)
        /// </summary>
        public float ForwardTestFactor = 0.75f;
        private Vector3[] _forwardTestOffsets;
        private Vector3[] ForwardTestOffsets
        {
            get
            {
                if (_forwardTestOffsets == null)
                {
                    if (ForwardTestFactor > 0)
                    {
                        _forwardTestOffsets = new Vector3[] {
                        new Vector3(0,0,0),
                        new Vector3(0,ForwardTestFactor * ColliderSize.y,0),
                        new Vector3(0,-ForwardTestFactor * ColliderSize.y,0),
                    };
                    }
                    else
                    {
                        _forwardTestOffsets = originOffsets;
                    }
                }
                return _forwardTestOffsets;
            }
        }
        
        public LayerMask Solid;
        public LayerMask OneWayPlatforms;
        

        protected abstract void OnTransitionToAir();
        protected abstract void OnGroundHit(RaycastHit2D hit);
        protected abstract void OnCeilingHit(RaycastHit2D hit);
        protected abstract void OnWallHit(RaycastHit2D hit, int Direction);
        protected abstract void OnCrushed();
         

        public void FixedUpdate()
        {
            Velocity.x = Mathf.Clamp(Velocity.x, -MaximumVelocity.x, MaximumVelocity.x);
            Velocity.y = Mathf.Clamp(Velocity.y, -MaximumVelocity.y, MaximumVelocity.y);

            this.transform.position += (Vector3)Velocity * Time.fixedDeltaTime;

            ForwardTest();
            var downHit = DownTest();
            var upHit = UpTest();
            if (downHit && upHit) OnCrushed();

            Velocity.x = Mathf.Clamp(Velocity.x, -MaximumVelocity.x, MaximumVelocity.x);
            Velocity.y = Mathf.Clamp(Velocity.y, -MaximumVelocity.y, MaximumVelocity.y);
        }
        private bool DownTest(float Scale = 1.1f)
        {
            if ((Velocity.y <= 0) && MultiRaycast(-Vector2.up, Solid | OneWayPlatforms, ColliderSize.y, Scale:Scale, Offsets: DownTestOffsets, c: Color.yellow))
            { 
                OnGroundHit(lastHitResult);                
                return true;
            }
            OnTransitionToAir();
            return false;
        }
        private bool UpTest()
        {
            if (Velocity.y < 0) return false; //Ignore unless we're moving upwards
            if (MultiRaycast(Vector2.up, Solid, ColliderSize.y, Offsets: DownTestOffsets, c: Color.cyan))
            {
                OnCeilingHit(lastHitResult);
                return true;
            }
            return false;
        }
        private bool ForwardTest()
        {
            if (Velocity.x < 0 && MultiRaycast(-this.transform.right, Solid, ColliderSize.x, Offsets: ForwardTestOffsets, c: Color.red)) //Only test Left if we're moving Left
            {
                OnWallHit(lastHitResult, 1);
                return true;
            }
            if (Velocity.x > 0 && MultiRaycast(this.transform.right, Solid, ColliderSize.x, Offsets: ForwardTestOffsets, c: Color.magenta)) //Only test Right if we're moving Right
            {
                OnWallHit(lastHitResult, -1);
                return true;
            }
            return false;
        }

        RaycastHit2D lastHitResult = new RaycastHit2D();
        private Vector3[] originOffsets = new[] { new Vector3(0, 0, 0) };
        private bool MultiRaycast(Vector2 Direction, LayerMask mask, float ColliderSizeF, float Scale = 1f, Vector3[] Offsets = null,Color? c = null)
        {
            if (Offsets == null) Offsets = originOffsets;
            foreach (var offset in Offsets)
            {
                lastHitResult = Physics2D.Raycast(this.transform.position + offset + (Vector3)ColliderOffset, Direction, ColliderSizeF * Scale, mask); 
                if (lastHitResult != null && lastHitResult.collider != null) return true;
            }
            return false;

        }
        void OnDrawGizmos()
        {
            Gizmos.color = Color.cyan;
            gizmoDrawOffsets(this.transform.right, ColliderSize.x, ForwardTestOffsets);
            Gizmos.color = Color.magenta;
            gizmoDrawOffsets(-Vector2.up, ColliderSize.y, DownTestOffsets);

        }
        void gizmoDrawOffsets(Vector3 Direction, float ColliderSizeF, Vector3[] Offsets)
        {
            Vector3 center = this.transform.position + (Vector3)ColliderOffset;
            foreach (var offset in Offsets)
            {
                Gizmos.DrawLine(center + offset,
                    center + offset + (Direction * ColliderSizeF)
                    );
                
            }

        }
    }

}