﻿using UnityEngine;
using System.Collections;
using UnityEngine.Events;

namespace com.geeitsomelaldy.platforming
{
    [RequireComponent(typeof(BaseCollider))]
    public class PlayerHealth : MonoBehaviour
    {
        [System.Serializable]
        public class CollisionReaction
        {
            [Tooltip("Set the player velocity to this when hit")]
            public Vector2 Velocity; 
            [Tooltip("GameObject to turn on and off when Mercy Invulnerability is active")]
            public GameObject GameObjectToFlash;
            [Tooltip("Flash the Sprite Renderer every (seconds)")]
            public float FlashTime = 0.2f;
            [Tooltip("Disable player input to the Platform controller (seconds)")]
            public float DisableControlsFor = 0.5f;

            internal void Flash(bool p)
            { 
                if (GameObjectToFlash != null) GameObjectToFlash.SetActive(p);
            }
        }
        private BaseCollider _bc;
        private BaseCollider baseCollider
        {
            get
            {
                if (_bc == null)
                {
                    _bc = this.GetComponent<BaseCollider>();
                }
                return _bc;
            }
        }

        [Tooltip("How much health the character has")]
        public int HP = 1;
        private  int initialHP;
        [Tooltip("How long to remain invulnerable after being hit (seconds)")]
        public float MercyInvulnerabilityTime = 5;
        [Tooltip("The layers that contain enemies")]
        public LayerMask EnemyLayers;
        public bool MercyInvulnerabilityOnDeath = true;
        [Tooltip("What the character does after being hit")]
        public CollisionReaction Reaction;
        [Tooltip("What the character does when dying")]
        public DeathAndRespawnEvents Death;

        public bool SetRespawnPointToInitialPosition = true;
        private bool isInvulnerable;
        private PlayerInput _playerInput;
        private PlayerInput playerInput
        {
            get
            {
                if (_playerInput == null) _playerInput = this.GetComponent<PlayerInput>();
                return _playerInput;
            }
        }

        private PlayerPlatformController _platformController;
        private PlayerPlatformController platformController
        {
            get
            {
                if (_platformController == null) _platformController = this.GetComponent<PlayerPlatformController>();
                return _platformController;
            }
        }
        public Vector2 RespawnPoint;
        public IntEvent HPChanged;
        public void Start()
        {
            initialHP = HP;
            if (SetRespawnPointToInitialPosition)
            {
                RespawnPoint = this.transform.position;
            }
        } 
        bool isDying = false;

        public void HealFor(int n)
        {
            if (n <= 0) return;
            HP += n;
            if (HP > initialHP) HP = initialHP;
            HPChanged.Invoke(HP);
        }
        public void Kill()
        {
            isInvulnerable = true;
            isDying = true;
            StartCoroutine(Death.Respawn(platformController.gameObject, RespawnPoint, () =>
            {
                isDying = false;
                isInvulnerable = false;
                HP = initialHP;
                HPChanged.Invoke(HP);
                if (MercyInvulnerabilityOnDeath)
                {
                    StartCoroutine(MercyInvulnerability());
                }
            }));
        }
        public void Update()
        {
            if (isDying) return;
            if (baseCollider.HasHit(EnemyLayers))
            {
                var ienemy = baseCollider.hit.collider.gameObject.GetComponentInParent<IEnemy>();
                if (ienemy != null)
                { 
                    if (ienemy.AttackFrom(this.transform.position, platformController.Velocity, DamageType.Stomp))
                    { 
                        platformController.Velocity.y = -platformController.Velocity.y;
                        return;
                    }
                }
                int damage = ienemy.DamageDealt();
                if (!isInvulnerable && damage > 0)
                {
                    HP -= damage;
                    HPChanged.Invoke(HP);
                }
                if (HP <= 0)
                {
                    Kill();
                    
                }
                else if (!isInvulnerable && damage > 0)
                {
                    StartCoroutine(MercyInvulnerability(false));
                    if (Reaction.Velocity.magnitude > 0)
                    {
                        platformController.Velocity = new Vector2(
                            -platformController.Direction * Reaction.Velocity.x,
                            Reaction.Velocity.y
                            );
                    }
                }
            }
            
        }
        private bool isFlashing = false;
        private IEnumerator MercyInvulnFlash()
        {
            isFlashing = true;
            bool isOn = true;
            while (isInvulnerable && isFlashing && Reaction.FlashTime > 0)
            {
                isOn = !isOn;
                Reaction.Flash(isOn);
                for (float t = 0; t < Reaction.FlashTime; t += Time.deltaTime)
                {
                    if (!isInvulnerable || !isFlashing) break;
                    yield return null;
                }
                if (!isInvulnerable || !isFlashing) break;
            }
            Reaction.Flash(true);
        }
        private IEnumerator DisableControls()
        {
            if (Reaction.DisableControlsFor > 0)
            {
                playerInput.enabled = false;
                yield return new WaitForSeconds(Reaction.DisableControlsFor);
                playerInput.enabled = true;
            }
        }
        private IEnumerator MercyInvulnerability(bool disableControls = true)
        {
            Debug.Log("MERCY INVULN START");
            if (MercyInvulnerabilityTime > 0)
            {
                isInvulnerable = true;
                StartCoroutine(MercyInvulnFlash());
                if (disableControls) StartCoroutine(DisableControls());
                yield return new WaitForSeconds(MercyInvulnerabilityTime);
                Reaction.Flash(true);
                isFlashing = false;
                yield return null;
                isInvulnerable = false;
                Debug.Log("MERCY INVULN END");
            }
        }
    }
}